# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import multiprocessing
from concurrent.futures import as_completed, Executor, ProcessPoolExecutor
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from pathspec import PathSpec, Pattern, RegexPattern
from pathspec.patterns.gitwildmatch import GitWildMatchPattern


T = TypeVar("T")
Excludes = Optional[List[str]]


EXECUTOR = None  # deprecated, will be removed by 2.0

INCLUDE_PATTERN: str = r".+\.pyi?$"
"""
Regex pattern used to match against filenames that should be included in results when
recursively walking directories. Any files not matching this regex will be ignored.
"""

ROOT_MARKERS: List[Path] = [Path("pyproject.toml"), Path(".git"), Path(".hg")]
"""
List of :class:`pathlib.Path` objects used to "mark" the root of project. This is used
by :func:`gitignore` for finding and reading a :file:`.gitignore` file from the project
root of a given path or directory.
"""


def project_root(path: Path) -> Path:
    """
    Find the project root, looking upward from the given path.

    Looks through all parent paths until either the root is reached, or a directory
    is found that contains any of :attr:`ROOT_MARKERS`.
    """
    real_path = path.resolve()

    parents = list(real_path.parents)
    if real_path.is_dir():
        parents.insert(0, real_path)

    for parent in parents:
        if any((parent / marker).exists() for marker in ROOT_MARKERS):
            return parent

    return parent


def pathspec(
    patterns: Excludes, *, style: Type[Pattern] = GitWildMatchPattern
) -> PathSpec:
    """
    Generate a `PathSpec` object for the given set of paths to include or exclude.

    If None is given, an empty PathSpec is returned.
    """

    if patterns:
        return PathSpec.from_lines(style, patterns)

    return PathSpec([])


def gitignore(path: Path) -> PathSpec:
    """
    Generate a `PathSpec` object for a .gitignore file in the given directory.

    If none is found, an empty PathSpec is returned. If the path is not a directory,
    `ValueError` is raised.
    """
    if not path.is_dir():
        raise ValueError(f"path {path} not a directory")

    gi_path = path / ".gitignore"

    lines: Excludes = None
    if gi_path.is_file():
        lines = gi_path.read_text().splitlines()

    return pathspec(lines, style=GitWildMatchPattern)


class Trailrunner:
    """
    Self-contained Trailrunner instance with configurable multiprocessing semantics.

    The basic API functions above are lightweight wrappers calling their respective
    methods on fresh instances of the :class:`Trailrunner` class with no arguments.

    By default, uses a process pool executor with "spawn" child processes, to enforce
    consistent behavior when running functions on Linux, macOS, and Windows. This can
    be overridden for each individual instance by passing an executor factory during
    initialization.
    """

    DEFAULT_EXECUTOR: Optional[Callable[[], Executor]] = None
    """
    Global override for the default executor, used when `executor_factory` is not
    passed to new instances. This is intended primarily for use in unit testing, when
    it is desirable to force using thread pools by default.

    Prefer explicitly passing `executor_factory` when possible to avoid conflicting
    behavior with other packages that also use trailrunner.
    """

    EXCLUDED: Dict[Path, str] = {}

    def __init__(
        self,
        *,
        concurrency: int = 0,
        context: Optional[multiprocessing.context.BaseContext] = None,
        executor_factory: Optional[Callable[[], Executor]] = None,
    ):
        """
        :param concurreny: maximum number of child processes to use for jobs.
            Values ``< 1`` will use the default concurrency level from
            :mod:`concurrent.futures.ProcessPoolExecutor`. Ignored if
            :attr:`DEFAULT_EXECUTOR` is set, or `executor_factory` is passed.
        :param context: :mod:`multiprocessing` context used by the default process pool
            executor. Ignored if :attr:`DEFAULT_EXECUTOR` is set, or `executor_factory`
            is passed.
        :param executor_factory: Alternative executor factory. Must be a function that
            takes no arguments, and returns an instance
            of :class:`concurrent.futures.Executor`.
        """
        self.concurrency = concurrency
        self.context = context or multiprocessing.get_context("spawn")
        self.executor_factory = (
            executor_factory or self.DEFAULT_EXECUTOR or self._default_executor
        )

    def _default_executor(self) -> Executor:
        """
        Default executor factory using a ProcessPoolExecutor.
        """
        if EXECUTOR:  # deprecated
            return EXECUTOR()
        concurrency = self.concurrency if self.concurrency > 0 else None
        return ProcessPoolExecutor(concurrency, mp_context=self.context)

    def walk(self, path: Path, *, excludes: Excludes = None) -> Iterator[Path]:
        """
        Generate all significant file paths, starting from the given path.

        Finds the project root and any associated gitignore. Filters any paths that match
        a gitignore pattern. Recurses into subdirectories, and otherwise only includes
        files that match the :attr:`trailrunner.core.INCLUDE_PATTERN` regex.

        Optional `excludes` parameter allows supplying an extra set of paths (or
        gitignore-style patterns) to exclude from the final results.

        Returns a generator that yields each significant file as the tree is walked.
        """
        root = project_root(path)
        ignore = gitignore(root) + pathspec(excludes)
        include = PathSpec([RegexPattern(INCLUDE_PATTERN)])

        def gen(children: Iterable[Path], *, explicit: bool = False) -> Iterator[Path]:
            for child in children:
                relative = child.resolve().relative_to(root)
                if ignore.match_file(relative):
                    self.EXCLUDED[child] = "matched gitignore"
                    continue

                if child.is_file() and (explicit or include.match_file(child)):
                    yield child

                elif child.is_dir():
                    yield from gen(child.iterdir())

        return gen([path], explicit=True)

    def run(self, paths: Iterable[Path], func: Callable[[Path], T]) -> Dict[Path, T]:
        """
        Run a given function once for each path, using an executor for concurrency.

        For each path given, `func` will be called with `path` as the only argument.
        To pass any other positional or keyword arguments, use `functools.partial`.

        Results from each path will be returned as a dictionary mapping path to result.
        """
        paths = list(paths)

        with self.executor_factory() as exe:
            results = list(exe.map(func, paths))

        return dict(zip(paths, results))

    def run_iter(
        self, paths: Iterable[Path], func: Callable[[Path], T]
    ) -> Generator[Tuple[Path, T], None, None]:
        """
        Run a given function once for each path, using an executor for concurrency.

        For each path given, `func` will be called with `path` as the only argument.
        To pass any other positional or keyword arguments, use `functools.partial`.

        Each path, and the function result, will be yielded as they are completed.
        """
        with self.executor_factory() as exe:
            futures = {exe.submit(func, path): path for path in paths}
            for future in as_completed(futures):
                path = futures[future]
                value = future.result()
                yield path, value

    def walk_and_run(
        self,
        paths: Iterable[Path],
        func: Callable[[Path], T],
        *,
        excludes: Excludes = None,
    ) -> Dict[Path, T]:
        """
        Walks each path given, and runs the given function on all gathered paths.

        See :meth:`Trailrunner.walk` for details on how paths are gathered, and
        :meth:`Trailrunner.run` for how functions are run for each gathered path.
        """
        all_paths: List[Path] = []
        for path in paths:
            all_paths.extend(self.walk(path, excludes=excludes))

        return self.run(all_paths, func)


# Maintain basic API with a default TrailRunner instance


def walk(path: Path, *, excludes: Excludes = None) -> Iterator[Path]:
    """
    Generate all significant file paths, starting from the given path.

    Finds the project root and any associated gitignore. Filters any paths that match
    a gitignore pattern. Recurses into subdirectories, and otherwise only includes
    files that match the :attr:`trailrunner.core.INCLUDE_PATTERN` regex.

    Optional `excludes` parameter allows supplying an extra set of paths (or
    gitignore-style patterns) to exclude from the final results.

    Returns a generator that yields each significant file as the tree is walked.
    """
    return Trailrunner().walk(path, excludes=excludes)


def run(paths: Iterable[Path], func: Callable[[Path], T]) -> Dict[Path, T]:
    """
    Run a given function once for each path, using a process pool for concurrency.

    For each path given, `func` will be called with `path` as the only argument.
    To pass any other positional or keyword arguments, use `functools.partial`.

    Results from each path will be returned as a dictionary mapping path to result.

    Uses a process pool with "spawned" processes that share no state with the parent
    process, to enforce consistent behavior on Linux, macOS, and Windows, where forked
    processes are not possible.
    """
    return Trailrunner().run(paths, func)


def run_iter(
    paths: Iterable[Path], func: Callable[[Path], T]
) -> Generator[Tuple[Path, T], None, None]:
    """
    Run a given function once for each path, using a process pool for concurrency.

    For each path given, `func` will be called with `path` as the only argument.
    To pass any other positional or keyword arguments, use `functools.partial`.

    Each path, and the function result, will be yielded as they are completed.

    Uses a process pool with "spawned" processes that share no state with the parent
    process, to enforce consistent behavior on Linux, macOS, and Windows, where forked
    processes are not possible.
    """
    return Trailrunner().run_iter(paths, func)


def walk_and_run(
    paths: Iterable[Path], func: Callable[[Path], T], *, excludes: Excludes = None
) -> Dict[Path, T]:
    """
    Walks each path given, and runs the given function on all gathered paths.

    See :func:`walk` for details on how paths are gathered, and :func:`run` for how
    functions are run for each gathered path.
    """
    return Trailrunner().walk_and_run(paths, func, excludes=excludes)



# Copyright Amethyst Reese
# Licensed under the MIT license

from __future__ import annotations

from pathlib import Path
from typing import Any

import click

try:
    print: Any
    from rich import print as print
except ImportError:
    from pprint import pprint as print

from . import core


@click.group()
def main() -> None:
    pass


@main.command()
@click.argument("path", required=True, type=click.Path(path_type=Path))
def debug(path: Path) -> None:
    """
    Output debug info about the given path
    """
    root = core.project_root(path)
    print(f"root = {root}")
    gitignore = core.gitignore(root)
    print(f"len(gitignore) = {len(gitignore.patterns)}")


@main.command()
@click.argument("path", required=True, type=click.Path(path_type=Path))
def walk(path: Path) -> None:
    """
    Print all paths found when walking
    """
    runner = core.Trailrunner()
    for child in runner.walk(path):
        print(child)

    print()
    for child, reason in runner.EXCLUDED.items():
        print(f"{child} excluded for {reason!r}")


if __name__ == "__main__":
    main()



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

"""
Run things on paths
"""

__author__ = "Amethyst Reese"
from .__version__ import __version__
from .core import (
    gitignore,
    project_root,
    run,
    run_iter,
    Trailrunner,
    walk,
    walk_and_run,
)

__all__ = [
    "__author__",
    "__version__",
    "gitignore",
    "project_root",
    "run",
    "run_iter",
    "Trailrunner",
    "walk",
    "walk_and_run",
]



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator, Iterator
from unittest import TestCase
from unittest.mock import Mock, patch

from pathspec import PathSpec
from pathspec.patterns.gitwildmatch import GitWildMatchPattern

from trailrunner import core


@contextmanager
def cd(path: Path) -> Iterator[None]:
    try:
        cwd = Path.cwd()
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


def getpid(path: Path) -> int:
    return os.getpid()


@patch.object(core.Trailrunner, "DEFAULT_EXECUTOR", ThreadPoolExecutor)
class CoreTest(TestCase):
    maxDiff = None

    def setUp(self) -> None:
        self.temp_dir = TemporaryDirectory()
        self.td = Path(self.temp_dir.name).resolve()

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_executor(self) -> None:
        def spawn_factory() -> ProcessPoolExecutor:
            return ProcessPoolExecutor(mp_context=multiprocessing.get_context("spawn"))

        parent = os.getpid()
        inputs = [Path()]
        expected = {
            Path(): parent,
        }

        with self.subTest("explicit thread pool"):
            result = core.Trailrunner(executor_factory=ThreadPoolExecutor).run(
                inputs, getpid
            )
            self.assertEqual(expected, result)

        with self.subTest("explicit spawn factory"):
            parent = os.getpid()
            results = core.Trailrunner(executor_factory=spawn_factory).run(
                inputs, getpid
            )
            self.assertNotEqual(expected, results)

        with self.subTest("patched thread pool"):
            result = core.Trailrunner().run(inputs, getpid)
            self.assertEqual(expected, result)

        with self.subTest("patched default is none"):
            with patch.object(core.Trailrunner, "DEFAULT_EXECUTOR", None):
                result = core.Trailrunner().run(inputs, getpid)
                self.assertNotEqual(expected, result)  # spawned process

                with self.subTest("deprecated EXECUTOR"):
                    with patch("trailrunner.core.EXECUTOR") as mock_exe:
                        core.Trailrunner().run(inputs, getpid)
                        mock_exe.assert_called_with()

    @patch("trailrunner.core.ProcessPoolExecutor")
    def test_concurrency(self, pool_mock: Mock) -> None:
        (self.td / "foo.py").write_text("\n")
        (self.td / "bar.py").write_text("\n")

        with patch.object(core.Trailrunner, "DEFAULT_EXECUTOR", None):
            for level, expected_param in [
                (1, 1),
                (5, 5),
                (0, None),
                (-365, None),
            ]:
                with self.subTest(level):
                    tr = core.Trailrunner(concurrency=level)
                    tr.walk_and_run([self.td], Path)
                    pool_mock.assert_called_with(expected_param, mp_context=tr.context)

    def test_project_root_empty(self) -> None:
        result = core.project_root(self.td)
        self.assertTrue(self.td.relative_to(result))

    def test_project_root_basic(self) -> None:
        (self.td / ".git").mkdir()
        (self.td / "foo.py").write_text("\n")
        (self.td / "frob").mkdir()
        (self.td / "frob" / "berry.py").write_text("\n")

        with self.subTest("root"):
            result = core.project_root(self.td)
            self.assertEqual(self.td, result)

        with self.subTest("root file"):
            result = core.project_root(self.td / "foo.py")
            self.assertEqual(self.td, result)

        with self.subTest("subdir"):
            result = core.project_root(self.td / "frob")
            self.assertEqual(self.td, result)

        with self.subTest("subdir file"):
            result = core.project_root(self.td / "frob/berry.py")
            self.assertEqual(self.td, result)

        with self.subTest("local root"):
            with cd(self.td):
                result = core.project_root(Path("frob"))
                self.assertEqual(self.td, result)

        with self.subTest("local subdir"):
            with cd(self.td / "frob"):
                result = core.project_root(Path("berry.py"))
                self.assertEqual(self.td, result)

    def test_project_root_multilevel(self) -> None:
        (self.td / ".hg").mkdir()
        inner = self.td / "foo" / "bar"
        inner.mkdir(parents=True)
        (inner / "pyproject.toml").write_text("\n")
        (inner / "fuzz").mkdir()
        (inner / "fuzz" / "ball.py").write_text("\n")

        with self.subTest("root"):
            result = core.project_root(self.td)
            self.assertEqual(self.td, result)

        with self.subTest("subdir"):
            result = core.project_root(self.td / "foo")
            self.assertEqual(self.td, result)

        with self.subTest("inner root"):
            result = core.project_root(inner)
            self.assertEqual(inner, result)

        with self.subTest("inner file"):
            result = core.project_root(inner / "fuzz" / "ball.py")
            self.assertEqual(inner, result)

    def test_gitignore(self) -> None:
        with self.subTest("no .gitignore"):
            result = core.gitignore(self.td)
            self.assertIsInstance(result, PathSpec)
            self.assertListEqual([], list(result.patterns))

        (self.td / "foo.py").write_text("\n")

        with self.subTest("path is file"):
            with self.assertRaisesRegex(ValueError, "path .+ not a directory"):
                core.gitignore(self.td / "foo.py")

        (self.td / ".gitignore").write_text("foo/\n*.c\n")

        with self.subTest("valid file"):
            result = core.gitignore(self.td)
            self.assertIsInstance(result, PathSpec)
            self.assertTrue(result.patterns)
            for pattern in result.patterns:
                self.assertIsInstance(pattern, GitWildMatchPattern)

    def test_walk(self) -> None:
        (self.td / ".git").mkdir()
        inner = self.td / "inner" / "subproject"
        inner.mkdir(parents=True)
        (inner / "pyproject.toml").write_text("\n")
        (inner / "requirements.txt").write_text("\n")
        (inner / "fuzz").mkdir()
        (inner / "fuzz" / "ball.py").write_text("\n")
        (inner / "fuzz" / "data.txt").write_text("\n")
        (self.td / "foo" / "bar").mkdir(parents=True)
        (self.td / "foo" / "a.py").write_text("\n")
        (self.td / "foo" / "bar" / "b.py").write_text("\n")
        (self.td / "foo" / "bar" / "c.pyi").write_text("\n")
        (self.td / "foo" / "d.cpp").write_text("\n")
        (self.td / "vendor" / "useful").mkdir(parents=True)
        (self.td / "vendor" / "useful" / "old.py").write_text("\n")

        with self.subTest("explicit txt files"):
            result = sorted(core.walk(inner / "requirements.txt"))
            expected = [inner / "requirements.txt"]
            self.assertListEqual(expected, result)

        with self.subTest("absolute root no gitignore"):
            result = sorted(core.walk(self.td))
            expected = [
                self.td / "foo" / "a.py",
                self.td / "foo" / "bar" / "b.py",
                self.td / "foo" / "bar" / "c.pyi",
                self.td / "inner" / "subproject" / "fuzz" / "ball.py",
                self.td / "vendor" / "useful" / "old.py",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("absolute subdir no gitignore"):
            result = sorted(core.walk(self.td / "foo"))
            expected = [
                self.td / "foo" / "a.py",
                self.td / "foo" / "bar" / "b.py",
                self.td / "foo" / "bar" / "c.pyi",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("absolute subdir no gitignore with excludes"):
            result = sorted(core.walk(self.td / "foo", excludes=["foo/bar/", "b.py"]))
            expected = [
                self.td / "foo" / "a.py",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("local root no gitignore"):
            with cd(self.td):
                result = sorted(core.walk(Path(".")))
                expected = [
                    Path("foo") / "a.py",
                    Path("foo") / "bar" / "b.py",
                    Path("foo") / "bar" / "c.pyi",
                    Path("inner") / "subproject" / "fuzz" / "ball.py",
                    Path("vendor") / "useful" / "old.py",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("local subdir no gitignore"):
            with cd(self.td):
                result = sorted(core.walk(Path("foo")))
                expected = [
                    Path("foo") / "a.py",
                    Path("foo") / "bar" / "b.py",
                    Path("foo") / "bar" / "c.pyi",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("local subdir no gitignore with excludes"):
            with cd(self.td):
                result = sorted(core.walk(Path("foo"), excludes=["foo/bar/"]))
                expected = [
                    Path("foo") / "a.py",
                ]
                self.assertListEqual(expected, result)

        (self.td / ".gitignore").write_text("vendor/\n*.pyi")

        with self.subTest("absolute root with gitignore"):
            result = sorted(core.walk(self.td))
            expected = [
                self.td / "foo" / "a.py",
                self.td / "foo" / "bar" / "b.py",
                self.td / "inner" / "subproject" / "fuzz" / "ball.py",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("absolute subdir with gitignore"):
            result = sorted(core.walk(self.td / "foo"))
            expected = [
                self.td / "foo" / "a.py",
                self.td / "foo" / "bar" / "b.py",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("absolute subdir with gitignore and excludes"):
            result = sorted(core.walk(self.td / "foo", excludes=["a.py"]))
            expected = [
                self.td / "foo" / "bar" / "b.py",
            ]
            self.assertListEqual(expected, result)

        with self.subTest("local root with gitignore"):
            with cd(self.td):
                result = sorted(core.walk(Path(".")))
                expected = [
                    Path("foo") / "a.py",
                    Path("foo") / "bar" / "b.py",
                    Path("inner") / "subproject" / "fuzz" / "ball.py",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("local subdir with gitignore"):
            with cd(self.td):
                result = sorted(core.walk(Path("foo")))
                expected = [
                    Path("foo") / "a.py",
                    Path("foo") / "bar" / "b.py",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("inner project snubs gitignore"):
            with cd(inner):
                result = sorted(core.walk(Path(".")))
                expected = [
                    Path("fuzz") / "ball.py",
                ]
                self.assertListEqual(expected, result)

    def test_walk_with_gitignore_matching_parents(self) -> None:
        """
        test gitignore that matches paths outside of project root
        """
        root = self.td / "inner"
        root.mkdir()
        (root / "pyproject.toml").write_text("")
        (root / "foo.py").write_text("")
        (root / "something.py").write_text("")
        (root / "inner").mkdir()
        (root / "inner" / "voice.py").write_text("")

        with self.subTest("without gitignore"):
            runner = core.Trailrunner()
            result = sorted(runner.walk(root))
            expected = sorted(
                [
                    root / "foo.py",
                    root / "something.py",
                    root / "inner" / "voice.py",
                ]
            )
            self.assertListEqual(expected, result)

        (root / ".gitignore").write_text("something.py\n")

        with self.subTest("exclude something"):
            runner = core.Trailrunner()
            result = sorted(runner.walk(root))
            expected = sorted(
                [
                    root / "foo.py",
                    root / "inner" / "voice.py",
                ]
            )
            self.assertListEqual(expected, result)

        with self.subTest("exclude something (relative)"):
            with cd(root):
                runner = core.Trailrunner()
                result = sorted(runner.walk(Path(".")))
                expected = sorted(
                    [
                        Path("foo.py"),
                        Path("inner") / "voice.py",
                    ]
                )
                self.assertListEqual(expected, result)

        with self.subTest("exclude something (relative inner)"):
            with cd(root / "inner"):
                runner = core.Trailrunner()
                result = sorted(runner.walk(Path(".")))
                expected = sorted(
                    [
                        Path("voice.py"),
                    ]
                )
                self.assertListEqual(expected, result)

        (root / ".gitignore").write_text("inner/\n")

        with self.subTest("exclude inner"):
            runner = core.Trailrunner()
            result = sorted(runner.walk(root))
            expected = sorted(
                [
                    root / "foo.py",
                    root / "something.py",
                ]
            )
            self.assertListEqual(expected, result)

        with self.subTest("exclude inner (relative)"):
            with cd(root):
                runner = core.Trailrunner()
                result = sorted(runner.walk(Path(".")))
                expected = sorted(
                    [
                        Path("foo.py"),
                        Path("something.py"),
                    ]
                )
                self.assertListEqual(expected, result)

        with self.subTest("exclude inner (relative inner)"):
            with cd(root / "inner"):
                runner = core.Trailrunner()
                result = sorted(runner.walk(Path(".")))
                expected = sorted([])
                self.assertListEqual(expected, result)

    def test_run(self) -> None:
        def get_posix(path: Path) -> str:
            return path.as_posix()

        paths = [
            Path("foo") / "bar" / "baz.py",
            Path("bingo.py"),
            Path("/frob/glob.pyi"),
        ]
        expected = {p: p.as_posix() for p in paths}
        result = core.run(paths, get_posix)
        self.assertDictEqual(expected, result)

    def test_run_iter(self) -> None:
        def get_posix(path: Path) -> str:
            return path.as_posix()

        paths = [
            Path("foo") / "bar" / "baz.py",
            Path("bingo.py"),
            Path("/frob/glob.pyi"),
        ]
        expected = {p: p.as_posix() for p in paths}

        gen = core.run_iter(paths, get_posix)
        self.assertIsInstance(gen, Generator)

        result = {}
        for path, value in gen:
            self.assertIn(path, paths)
            result[path] = value

        self.assertDictEqual(expected, result)

    def test_walk_then_run(self) -> None:
        (self.td / "pyproject.toml").write_text("\n")
        (self.td / "foo").mkdir()
        (self.td / "foo" / "foo.py").write_text("\n")
        (self.td / "foo" / "bar.py").write_text("\n")
        (self.td / "foo" / "car.py").write_text("\n")
        (self.td / "foo" / "car.pyi").write_text("\n")
        (self.td / "foo" / "data.dat").write_text("\n")
        (self.td / "vendor").mkdir()
        (self.td / "vendor" / "something.py").write_text("\n")
        (self.td / "vendor" / "everything.py").write_text("\n")

        def say_hello(path: Path) -> str:
            return f"hello {path}"

        with self.subTest("local root no gitignore"):
            with cd(self.td):
                result = sorted(core.walk_and_run([Path(".")], say_hello).keys())
                expected = [
                    Path("foo") / "bar.py",
                    Path("foo") / "car.py",
                    Path("foo") / "car.pyi",
                    Path("foo") / "foo.py",
                    Path("vendor") / "everything.py",
                    Path("vendor") / "something.py",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("local subdir no gitignore"):
            with cd(self.td):
                result = sorted(core.walk_and_run([Path("foo")], say_hello).keys())
                expected = [
                    Path("foo") / "bar.py",
                    Path("foo") / "car.py",
                    Path("foo") / "car.pyi",
                    Path("foo") / "foo.py",
                ]
                self.assertListEqual(expected, result)

        (self.td / ".gitignore").write_text("**/foo.py\nvendor/\n")

        with self.subTest("local root with gitignore"):
            with cd(self.td):
                result = sorted(core.walk_and_run([Path(".")], say_hello).keys())
                expected = [
                    Path("foo") / "bar.py",
                    Path("foo") / "car.py",
                    Path("foo") / "car.pyi",
                ]
                self.assertListEqual(expected, result)

        with self.subTest("local subdir with gitignore"):
            with cd(self.td):
                result = sorted(core.walk_and_run([Path("foo")], say_hello).keys())
                expected = [
                    Path("foo") / "bar.py",
                    Path("foo") / "car.py",
                    Path("foo") / "car.pyi",
                ]
                self.assertListEqual(expected, result)
